import paramiko
from scp import SCPClient


class Transport:

    def __init__(self, server, port, user, password):
        self.ssh = self._createSSHClient(server=server, port=port, user=user, password=password)
        self.scp = SCPClient(self.ssh.get_transport())

    @staticmethod
    def _createSSHClient(server, port, user, password):
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(server, port, user, password)
        return client





import inspect
import logging
import os
import re
import subprocess
import sys
import time
import traceback
from typing import Dict, Union, Tuple, Optional, Any

from appium_extended_helpers.helpers_decorators import log_debug
# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(
#    __file__))))  # The sys.path.append line adds the parent directory of the tests directory to the Python module search path, allowing you to import modules from the root folder.

from appium_extended_utils import operations

logger = logging.getLogger(__name__)


class Adb:

    @staticmethod
    def get_device_uuid() -> Union[str, None]:
        """
        Получает UUID подключенного устройства Android с помощью команды adb.
        Returns:
            UUID в виде строки.
        """
        logger.debug("get_device_uuid()")

        # Определение команды для выполнения с помощью adb для получения списка устройств
        command = ['adb', 'devices']

        try:
            # Выполнение команды и получение вывода
            response = str(subprocess.check_output(command))

            # Извлечение списка устройств из полученного вывода с использованием регулярных выражений
            device_list = re.findall(r'(\d+\.\d+\.\d+\.\d+:\d+|\d+)', response)

            try:
                # Возвращение первого устройства из списка (UUID подключенного устройства Android)
                logger.debug(f"get_device_uuid() > {device_list[0]}")

                return device_list[0]
            except IndexError:
                logger.error("get_device_uuid() > None")
                logger.error("Нет подключенных устройств")
                return None
        except subprocess.CalledProcessError as e:
            logger.error("get_device_uuid() > None")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return None

    @staticmethod
    def get_device_model() -> Optional[str]:
        """
        Получает модель подключенного устройства Android с помощью команды adb.
        Возвращает модель устройства.
        """
        logger.debug("get_device_model()")

        command = ["adb", "shell", "getprop", "ro.product.model"]
        try:
            # Выполнение команды и получение вывода
            model = subprocess.check_output(command)
            # Преобразование байтовой строки в обычную строку и удаление пробельных символов и символов перевода строки
            model = model.decode().strip()
            return model

        except subprocess.CalledProcessError as e:
            logger.error("get_device_model() > None")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)

    @staticmethod
    def push(source: str, destination: str) -> bool:
        """
        Копирует файл или директорию на подключенное устройство.

        Аргументы:
            source (str): Путь к копируемому файлу или директории.
            destination (str): Путь назначения на устройстве.

        Возвращает:
            bool: True, если файл или директория были успешно скопированы, False в противном случае.
        """
        logger.debug(f"push() < {source=}, {destination=}")

        if not os.path.exists(source):
            logger.error(f"Путь к копируемому файлу или директории не существует: {source=}")
            return False

        command = ["adb", "push", source, destination]
        try:
            subprocess.run(command, check=True)
            logger.debug("push() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("push() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def pull(source: str, destination: str) -> bool:
        """
        Копирует файл или директорию с подключенного устройства.

        Аргументы:
            source (str): Путь к исходному файлу или директории на устройстве.
            destination (str): Целевой путь для сохранения скопированного файла или директории.

        Возвращает:
            bool: True, если файл или директория были успешно скопированы, False в противном случае.
        """
        logger.debug(f"pull() < {source=}, {destination=}")

        command = ["adb", "pull", source, destination]
        try:
            subprocess.run(command, check=True)
            logger.debug("pull() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("pull() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def install_app(source: str) -> bool:
        """
        Устанавливает файл APK на подключенном устройстве.

        Аргументы:
            source (str): Путь к файлу APK для установки.

        Возвращает:
            bool: True, если файл APK был успешно установлен, False в противном случае.
        """
        logger.debug(f"install() < {source=}")

        command = ["adb", "install", "-r", source]
        try:
            subprocess.run(command, check=True)
            logger.debug("install() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("install() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def is_app_installed(package) -> bool:
        """
        Проверяет, установлен ли пакет.
        """
        logger.debug(f"is_installed() < {package=}")

        command = "adb shell pm list packages"
        try:
            result = subprocess.check_output(command, shell=True).decode().strip()
            # Фильтруем пакеты
            if any([line.strip().endswith(package) for line in result.splitlines()]):
                logger.debug("install() > True")
                return True
            logger.debug("install() > False")
            return False
        except subprocess.CalledProcessError as e:
            logger.error("install() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def uninstall_app(package: str) -> bool:
        """
        Удаляет указанный пакет с помощью ADB.

        Аргументы:
            package (str): Название пакета приложения для удаления.

        Возвращает:
            bool: True, если приложение успешно удалено, False в противном случае.
        """
        logger.debug(f"uninstall_app() < {package=}")

        command = ['adb', 'uninstall', package]
        try:
            subprocess.run(command, check=True)
            logger.debug("uninstall_app() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("uninstall_app() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def start_activity(package: str, activity: str) -> bool:
        """
        Запускает активность на подключенном устройстве.

        Аргументы:
            package (str): Название пакета активности.
            activity (str): Название запускаемой активности.

        Возвращает:
            bool: True, если активность была успешно запущена, False в противном случае.
        """
        logger.debug(f"start_activity() < {package=}, {activity=}")

        command = ['adb', 'shell', 'am', 'start', '-n', f'{package}/{activity}']
        try:
            subprocess.check_output(command)
            logger.debug("start_activity() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("start_activity() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def get_current_activity() -> Union[str, None]:
        """
        Получает активити текущего запущенного приложения на устройстве с помощью ADB.
        Возвращает имя активити в виде строки или None, если произошла ошибка.

        Возвращает:
            str: Название активити текущего запущенного приложения, либо None, если произошла ошибка.
        """

        # Вывод информации о запуске функции в лог
        logger.debug("get_current_activity()")

        # Команда для ADB для получения информации о текущих окнах
        command = ['adb', 'shell', 'dumpsys', 'window', 'windows']

        try:
            # Выполнение команды и декодирование результата
            result = subprocess.check_output(command, shell=True).decode().strip()

            # Определение паттерна для поиска нужной информации в результатах
            pattern = r'mCurrentFocus|mFocusedApp'

            # Вызов функции grep_pattern для поиска соответствия паттерну
            matched_lines = operations.grep_pattern(input_string=result, pattern=pattern)

            # Если были найдены соответствующие строки
            if matched_lines:
                for line in matched_lines:
                    # Поиск имени активити в строке
                    match = re.search(r'\/([^\/}]*)', line)
                    if match:
                        # Возвращаем найденное значение, исключая '/'
                        activity_name = match.group(1)
                        logger.debug(f"get_current_activity() > {activity_name}")
                        return activity_name

            # Если не удалось найти активити, возвращаем None
            logger.error("get_current_activity() > None")
            return None
        except subprocess.CalledProcessError as e:
            # Обработка ошибки при выполнении команды
            logger.error(e)

            # Вывод информации о трассировке в лог
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)

            logger.error("get_current_activity() > None")
            return None

    @staticmethod
    def get_current_package() -> Union[str, None]:
        """
        Получает пакет текущего запущенного приложения на устройстве с помощью ADB.
        Возвращает имя пакета в виде строки или None, если произошла ошибка.

        Возвращает:
            str: Название пакета текущего запущенного приложения, либо None, если произошла ошибка.
        """
        # Вывод информации о запуске функции в лог
        logger.debug("get_current_app_package()")

        # Команда для ADB для получения информации о текущих окнах
        command = ['adb', 'shell', 'dumpsys', 'window', 'windows']

        try:
            # Выполнение команды и декодирование результата
            result = subprocess.check_output(command, shell=True).decode().strip()

            # Определение паттерна для поиска нужной информации в результатах
            pattern = r'mCurrentFocus|mFocusedApp'

            # Вызов функции grep_pattern для поиска соответствия паттерну
            matched_lines = operations.grep_pattern(input_string=result, pattern=pattern)

            # Если были найдены соответствующие строки
            if matched_lines:
                for line in matched_lines:
                    # Поиск имени пакета в строке
                    match = re.search(r'u0\s(.+?)/', line)
                    if match:
                        # Возвращаем найденное значение
                        package_name = match.group(1)
                        logger.debug(f"get_current_app_package() > {package_name}")
                        return package_name

            # Если не удалось найти имя пакета, возвращаем None
            logger.error("get_current_app_package() > None")
            return None
        except subprocess.CalledProcessError as e:
            # Обработка ошибки при выполнении команды
            logger.error(e)

            # Вывод информации о трассировке в лог
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)

            logger.error("get_current_app_package() > None")
            return None

    @staticmethod
    def close_app(package: str) -> bool:
        """
        Принудительно останавливает указанный пакет с помощью ADB.
    
        Аргументы:
            package (str): Название пакета приложения для закрытия.
    
        Возвращает:
            bool: True, если приложение успешно закрыто, False в противном случае.
        """
        logger.debug(f"close_app() < {package=}")

        command = ['adb', 'shell', 'am', 'force-stop', package]
        try:
            subprocess.run(command, check=True)
            logger.debug("close_app() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("close_app() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def reboot_app(package: str, activity: str) -> bool:
        """
        Перезапускает приложение, закрывая его и затем запуская указанную активность.

        Аргументы:
            package (str): Название пакета приложения.
            activity (str): Название активности для запуска.

        Возвращает:
            bool: True, если перезапуск приложения выполнен успешно, False в противном случае.
        """
        logger.debug(f"reboot_app() < {package=}, {activity=}")

        # Закрытие приложения
        if not Adb.close_app(package=package):
            logger.error("reboot_app() > False")
            return False

        # Запуск указанной активности
        if not Adb.start_activity(package=package, activity=activity):
            logger.error("reboot_app() > False")
            return False
        logger.debug("reboot_app() > True")
        return True

    @staticmethod
    def press_home() -> bool:
        """
        Отправляет событие нажатия кнопки Home на устройство с помощью ADB.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        logger.debug("press_home()")

        command = ['adb', 'shell', 'input', 'keyevent', 'KEYCODE_HOME']
        try:
            subprocess.run(command, check=True)
            logger.debug("press_home() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("press_home() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def press_back() -> bool:
        """
        Отправляет событие нажатия кнопки Back на устройство с помощью ADB.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """

        logger.debug("press_back()")

        command = ['adb', 'shell', 'input', 'keyevent', 'KEYCODE_BACK']
        try:
            subprocess.run(command, check=True)
            logger.debug("press_back() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("press_back() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def press_menu() -> bool:
        """
        Отправляет событие нажатия кнопки Menu на устройство с помощью ADB.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """

        logger.debug("press_menu()")

        command = ['adb', 'shell', 'input', 'keyevent', 'KEYCODE_MENU']
        try:
            subprocess.run(command, check=True)
            logger.debug("press_menu() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("adb.press_menu() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def input_keycode_num_(num: int) -> bool:
        """
        Отправляет событие нажатия клавиши с числовым значением на устройство с помощью ADB.
        Допустимые значения: 0-9, ADD, COMMA, DIVIDE, DOT, ENTER, EQUALS

        Аргументы:
            num (int): Числовое значение клавиши для нажатия.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """

        logger.debug(f"input_keycode_num_() < {num=}")

        command = ['adb', 'shell', 'input', 'keyevent', f'KEYCODE_NUMPAD_{num}']
        try:
            subprocess.run(command, check=True)
            logger.debug("input_keycode_num_() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("input_keycode_num_() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def input_keycode(keycode: str) -> bool:
        """
        Вводит указанный код клавиши на устройстве с помощью ADB.

        Аргументы:
            keycode (str): Код клавиши для ввода.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """

        logger.debug(f"input_keycode() < {keycode=}")

        command = ['adb', 'shell', 'input', 'keyevent', f'{keycode}']
        try:
            subprocess.run(command, check=True)
            logger.debug("input_keycode() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("input_keycode() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def input_by_virtual_keyboard(text: str, keyboard: Dict[str, tuple]) -> bool:
        """
        Вводит строку символов с помощью виртуальной клавиатуры.

        Аргументы:
            key (str): Строка символов для ввода.
            keyboard (dict): Словарь с маппингом символов на координаты нажатий.

        Возвращает:
            bool: True, если ввод выполнен успешно, False в противном случае.
        """
        logger.debug(f"input_by_virtual_keyboard() < {text=}, {keyboard=}")

        try:
            for char in text:
                # Вызываем функцию tap с координатами, соответствующими символу char
                Adb.tap(*keyboard[char])
            logger.debug("input_by_virtual_keyboard() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("input_by_virtual_keyboard() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def input_text(text: str) -> bool:
        """
        Вводит указанный текст на устройстве с помощью ADB.

        Аргументы:
            text (str): Текст для ввода.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """

        logger.debug(f"input_text() < {text=}")

        # Формируем команду для ввода текста с использованием ADB
        command = ['adb', 'shell', 'input', 'text', text]
        try:
            # Выполняем команду
            subprocess.run(command, check=True)
            logger.debug("input_text() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("input_text() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def tap(x: Union[str, int], y: Union[str, int]) -> bool:
        """
        Выполняет нажатие на указанные координаты на устройстве с помощью ADB.

        Аргументы:
            x: Координата X для нажатия.
            y: Координата Y для нажатия.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """

        logger.debug(f"tap() < {x=}, {y=}")

        # Формируем команду для выполнения нажатия по указанным координатам с использованием ADB
        command = ['adb', 'shell', 'input', 'tap', str(x), str(y)]
        try:
            subprocess.run(command, check=True)
            logger.debug("tap() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("tap() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def swipe(start_x: Union[str, int], start_y: Union[str, int],
              end_x: Union[str, int], end_y: Union[str, int],
              duration: int = 300) -> bool:
        """
        Выполняет свайп (перетаскивание) с одной точки на экране в другую на устройстве с помощью ADB.

        Аргументы:
            start_x: Координата X начальной точки свайпа.
            start_y: Координата Y начальной точки свайпа.
            end_x: Координата X конечной точки свайпа.
            end_y: Координата Y конечной точки свайпа.
            duration (int): Длительность свайпа в миллисекундах (по умолчанию 300).

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """

        logger.debug(f"swipe() < {start_x=}, {start_y=}, {end_x=}, {end_y=}, {duration=}")

        # Формируем команду для выполнения свайпа с использованием ADB
        command = ['adb', 'shell', 'input', 'swipe', str(start_x), str(start_y), str(end_x), str(end_y), str(duration)]
        try:
            # Выполняем команду
            subprocess.run(command, check=True)
            logger.debug("swipe() > True")
            return True
        except subprocess.CalledProcessError as e:
            # Логируем ошибку, если возникло исключение
            logger.error("swipe() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def check_vpn(ip_address: str = '') -> bool:
        """
        Проверяет, активно ли VPN-соединение на устройстве с помощью ADB.

        Аргументы:
            ip_address (str): IP-адрес для проверки VPN-соединения. Если не указан, используется значение из конфигурации.

        Возвращает:
            bool: True, если VPN-соединение активно, False в противном случае.
        """
        logger.debug(f"check_vpn() < {ip_address=}")

        # Определяем команду в виде строки
        command = "adb shell netstat"
        try:
            # Выполняем команду и получаем вывод
            output = subprocess.run(command, shell=True, capture_output=True, text=True, check=True)

            # Поиск строки
            lines = output.stdout.split("\n")
            for line in lines:
                if "ESTABLISHED" in line and ip_address in line:
                    logger.debug("check_vpn() True")
                    return True
            logger.debug("check_vpn() False")
            return False
        except subprocess.CalledProcessError as e:
            # Логируем ошибку, если возникло исключение
            logger.error("check_vpn() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def stop_logcat() -> bool:
        """
        Останавливает выполнение logcat на устройстве с помощью ADB.

        Возвращает:
            bool: True, если выполнение logcat остановлено успешно, False в противном случае.
        """
        logger.debug("stop_logcat()")
        if Adb.is_process_exist(name='logcat'):
            if Adb.kill_all(name='logcat'):
                logger.debug("stop_logcat() > True")
                return True
        logger.error("stop_logcat() > False")
        logger.debug("stop_logcat() [Запущенного процесса logcat не обнаружено]")
        return False

    @staticmethod
    def is_process_exist(name) -> bool:
        """
        Проверяет, запущен ли процесс, используя adb shell ps.

        Параметры:
            name (str): Имя процесса.

        Возвращает:
            bool: True если процесс с указанным именем существует, False в ином случае.
        """
        logger.debug(f"is_process_exist() < {name=}")
        command = ['adb', 'shell', 'ps']
        try:
            processes = subprocess.check_output(command, shell=True).decode().strip()
        except subprocess.CalledProcessError as e:
            logger.error("know_pid() > None")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        # Разделение вывода на строки и удаление пустых строк
        lines = processes.strip().split('\n')
        # Проход по каждой строке вывода, начиная с 2-й строки, игнорируя заголовки
        for line in lines[1:]:
            # Разделение строки на столбцы по пробелам
            columns = line.split()
            # Проверка, что строка имеет не менее 9 столбцов
            if len(columns) >= 9:
                # Извлечение PID и имени процесса из соответствующих столбцов
                _, process_name = columns[1], columns[8]
                # Сравнение имени процесса с искомым именем
                if name == process_name:
                    logger.debug("is_process_exist() > True")
                    return True
        # Возврат None, если процесс с заданным именем не найден
        logger.debug("is_process_exist() > False")
        return False

    @staticmethod
    def run_background_process(command: str, process: str = "") -> bool:
        """
        Запускает процесс в фоновом режиме на устройстве Android с использованием ADB.

        Аргументы:
            command (str): Команда для выполнения на устройстве.
            process (str): Название процесса, который будет запущен. По умолчанию "".
            Если process == "", то не будет проверяться его запуск в системе.

        Возвращает:
            bool: True, если процесс был успешно запущен, False в противном случае.
        """

        logger.debug(f"run_background_process() < {command=}")

        command = f"{command} nohup > /dev/null 2>&1 &"
        try:
            subprocess.Popen(command, stdout=subprocess.DEVNULL)  # не добавлять with
            if process != "":
                time.sleep(1)
                if not Adb.is_process_exist(name=process):
                    return False
            logger.debug("run_background_process() > True")
            return True
        except subprocess.CalledProcessError as e:
            logger.error("run_background_process() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

    @staticmethod
    def reload_adb() -> bool:
        """
        Перезапускает adb-сервер на устройстве.

        Возвращает:
            bool: True, если adb-сервер успешно перезапущен, False в противном случае.
        """
        logger.debug("reload_adb()")

        try:
            command = ['adb', 'kill-server']
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("reload_adb() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        # Ожидаем некоторое время перед запуском adb-сервера
        time.sleep(3)
        try:
            command = ['adb', 'start-server']
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("reload_adb() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        logger.debug("reload_adb() > True")
        return True

    @staticmethod
    def know_pid(name: str) -> Union[int, None]:
        """
        Находит Process ID (PID) процесса по его имени, используя adb shell ps.

        Параметры:
            name (str): Имя процесса, PID которого нужно найти.

        Возвращает:
            Union[int, None]: PID процесса, если он найден, или None, если процесс не найден.
        """

        logger.debug(f"know_pid() < {name=}")
        command = ['adb', 'shell', 'ps']
        try:
            processes = subprocess.check_output(command, shell=True).decode().strip()
        except subprocess.CalledProcessError as e:
            logger.error("know_pid() > None")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return None
        # Разделение вывода на строки и удаление пустых строк
        lines = processes.strip().split('\n')
        # Проход по каждой строке вывода, начиная с 2-й строки, игнорируя заголовки
        for line in lines[1:]:
            # Разделение строки на столбцы по пробелам
            columns = line.split()
            # Проверка, что строка имеет не менее 9 столбцов
            if len(columns) >= 9:
                # Извлечение PID и имени процесса из соответствующих столбцов
                pid, process_name = columns[1], columns[8]
                # Сравнение имени процесса с искомым именем
                if name == process_name:
                    logger.debug(f"know_pid() > {pid=}")
                    return int(pid)
        # Возврат None, если процесс с заданным именем не найден
        logger.error("know_pid() > None")
        logger.error("know_pid() [Процесс не обнаружен]")
        return None

    @staticmethod
    def kill_by_pid(pid: Union[str, int]) -> bool:
        """
        Отправляет сигнал SIGINT для остановки процесса по указанному идентификатору PID с помощью ADB.

        Аргументы:
            pid (str): Идентификатор PID процесса для остановки.

        Возвращает:
            bool: True, если процесс успешно остановлен, False в противном случае.
        """

        logger.debug(f"kill_by_pid() < {pid=}")

        command = ['adb', 'shell', 'kill', '-s', 'SIGINT', str(pid)]
        try:
            subprocess.call(command)
        except subprocess.CalledProcessError as e:
            logger.error("kill_by_pid() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        logger.debug("kill_by_pid() > True")
        return True

    @staticmethod
    def kill_by_name(name: str) -> bool:
        """
        Останавливает все процессы с указанным именем на устройстве с помощью ADB.

        Аргументы:
            name (str): Имя процесса для остановки.

        Возвращает:
            bool: True, если все процессы успешно остановлены, False в противном случае.
        """

        logger.debug(f"kill_by_name() < {name=}")

        command = ['adb', 'shell', 'pkill', '-l', 'SIGINT', str(name)]
        try:
            subprocess.call(command)
        except subprocess.CalledProcessError as e:
            logger.error("kill_by_name() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        logger.debug("kill_by_name() > True")
        return True

    @staticmethod
    def kill_all(name: str) -> bool:
        """
        Останавливает все процессы, соответствующие указанному имени, на устройстве с помощью ADB.

        Аргументы:
            name (str): Имя процесса или шаблон имени для остановки.

        Возвращает:
            bool: True, если все процессы успешно остановлены, False в противном случае.
        """

        logger.debug(f"kill_all() < {name=}")

        command = ['adb', 'shell', 'pkill', '-f', str(name)]
        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("kill_all() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        logger.debug("kill_all() > True")
        return True

    @staticmethod
    def delete_files_from_internal_storage(path: str) -> bool:
        """
        Удаляет файлы из внутреннего хранилища устройства с помощью ADB.

        Аргументы:
            path (str): Путь к файлам для удаления.

        Возвращает:
            bool: True, если файлы успешно удалены, False в противном случае.
        """

        logger.debug(f"delete_files_from_internal_storage() < {path=}")

        command = ['adb', 'shell', 'rm', '-rf', f'{path}*']
        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("delete_files_from_internal_storage() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        logger.debug("delete_files_from_internal_storage() > True")
        return True

    @staticmethod
    def pull_video(source: str = None, destination: str = ".", delete: bool = True) -> bool:
        """
        Копирует видеофайлы с устройства на компьютер с помощью ADB.

        Аргументы:
            wherefrom (str): Путь к исходным видеофайлам на устройстве.
            destination (str): Путь для сохранения скопированных видеофайлов.
            delete (bool): Удалять исходные видеофайлы с устройства после копирования (по умолчанию True).

        Возвращает:
            bool: True, если видеофайлы успешно скопированы, False в противном случае.
        """

        logger.debug(f"pull_video() < {destination=}")

        if not source:
            source = '/sdcard/Movies/'
        if source.endswith('/'):
            source = source + "/"
        if destination.endswith('/'):
            destination = destination + "/"

        command = ['adb', 'pull', f'{source}', f'{destination}']
        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("pull_video() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False

        if delete:
            command = ['adb', 'shell', 'rm', '-rf', f'{source}*']
            try:
                subprocess.run(command, check=True)
            except subprocess.CalledProcessError as e:
                logger.error("pull_video() > False")
                logger.error(e)
                traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
                logger.error(traceback_info)
                return False

            logger.debug("pull_video() > True")
        return True

    @staticmethod
    def stop_video() -> bool:
        """
        Останавливает запись видео на устройстве с помощью ADB.

        Возвращает:
            bool: True, если запись видео успешно остановлена, False в противном случае.
        """

        logger.debug("stop_video()")

        command = ['adb', 'shell', 'pkill', '-l', 'SIGINT', 'screenrecord']
        try:
            subprocess.call(command)
        except subprocess.CalledProcessError as e:
            logger.error("stop_video() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        logger.debug("stop_video() > True")
        return True

    @staticmethod
    def record_video(path: str = "sdcard/Movies/", filename: str = "screenrecord.mp4") -> \
            Union[subprocess.Popen[bytes], subprocess.Popen[Union[Union[str, bytes], Any]]]:
        """
        Записывает видео на устройстве с помощью ADB.

        Аргументы:
            path (str): Путь куда сохранить файл
            filename (str): Имя файла для сохранения видео.

        Возвращает:
            subprocess.CompletedProcess: Процесс записи видео.
        """

        logger.debug(f"record_video() < {filename}")
        if path.endswith('/'):
            path = path[:-1]
        if filename.endswith('.mp4'):
            filename = filename + ".mp4"

        command = ['adb', 'shell', 'screenrecord', f'{path}/{filename}']
        try:
            # Запускаем команду adb shell screenrecord для начала записи видео
            return subprocess.Popen(command)
        except subprocess.CalledProcessError as e:
            # Если произошла ошибка при выполнении команды, логируем ошибку и возвращаем False
            logger.error("record_video() > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)

    @staticmethod
    def start_record_video(path: str = "sdcard/Movies/", filename: str = "screenrecord.mp4") -> bool:
        """
        Отправляет команду на устройство для начала записи видео.

        Аргументы:
            path (str): Путь куда сохранить файл
            filename (str): Имя файла для сохранения видео.

        Возвращает:
            bool: True, если запись видео успешно начата, False в противном случае.
        """
        if path.endswith('/'):
            path = path[:-1]
        if not filename.endswith('.mp4'):
            filename = filename + ".mp4"

        command = ['adb', 'shell', 'screenrecord', f'{path}/{filename}']
        try:
            # Запускаем команду adb shell screenrecord для начала записи видео
            subprocess.Popen(command)  # не добавлять with
            return True
        except subprocess.CalledProcessError:
            # Если произошла ошибка при выполнении команды, возвращаем False
            return False

    @staticmethod
    def reboot() -> bool:
        """
        Перезагружает устройство с помощью ADB.

        Возвращает:
            bool: True, если перезагрузка успешно запущена, False в противном случае.
        """

        logger.debug("reboot()")

        command = ['adb', 'shell', 'reboot']
        try:
            subprocess.call(command)
        except subprocess.CalledProcessError as e:
            logger.error("reboot > False")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
            return False
        logger.debug("reboot() > True")
        return True

    @staticmethod
    def get_screen_resolution() -> Union[Tuple[int, int], None]:
        """
        Возвращает разрешение экрана устройства с помощью ADB.

        Возвращает:
            tuple[int, int] or None: Кортеж с шириной и высотой экрана в пикселях, или None в случае ошибки.
        """

        logger.debug("get_screen_resolution()")

        command = ['adb', 'shell', 'wm', 'size']
        try:
            output = subprocess.check_output(command).decode()
            if "Physical size" in output:
                resolution_str = output.split(":")[1].strip()
                width, height = resolution_str.split("x")
                logger.debug(f"get_screen_resolution() > {width=}, {height=}")
                return int(width), int(height)
            logger.error(f"Unexpected output from adb: {output}")
        except (subprocess.CalledProcessError, ValueError) as e:
            logger.error("get_screen_resolution() > None")
            logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            logger.error(traceback_info)
        return None

    def get_packages_list(self) -> list:
        packages_raw = self.execute(command="shell pm list packages")
        # Используем регулярное выражение для удаления "package:" из каждой строки
        packages_raw = re.sub(r'package:', '', packages_raw)
        # Разбиваем строки на список и удаляем пустые элементы
        packages_list = [package.strip() for package in packages_raw.split('\n') if package.strip()]
        return packages_list

    @staticmethod
    def execute(command: str):
        logger.debug(f"execute() < {command}")
        execute_command = ['adb', *command.split()]
        return subprocess.check_output(execute_command).decode()






# coding: utf-8
import base64
import logging
import os
import time
from typing import Union, List, Tuple, Optional, Dict

import cv2
import numpy as np
from PIL import Image
from pytesseract import pytesseract

from appium_extended_helpers import helpers_decorators
from appium_extended_terminal.terminal import Terminal
from appium_extended_utils import utils


class AppiumImage:
    """
    Класс работы с Appium.
    Обеспечивает работу с изображениями
    """

    def __init__(self, base):
        self.logger = base.logger
        self.driver = base.driver
        self.terminal = base.terminal

    def find_and_tap_in_drop_down_menu(self, item: str, items: List[str], drop_down_locator=None, timeout: int = 390):
        swipe_direction_down = True
        while True:
            try:
                item = item.lower()
                coord = self.get_text_coordinates(text=items)
                if item in coord.keys():
                    element_coord = utils.calculate_center_of_coordinates(coord[item])
                    self.terminal.tap(x=element_coord[0], y=element_coord[1])
                    return True

                # сортировка по последнему значению координат
                sorted_list = [item for item in sorted(coord.items(), key=lambda x: x[1][-1])]
                sorted_dicts = [dict({key: value}) for key, value in sorted_list]

                top_dict = sorted_dicts[1]  # Верхний словарь в списке
                bottom_dict = sorted_dicts[3]  # Нижний словарь в списке

                top_key = list(top_dict.keys())[0]  # Ключ первого словаря
                top_value = coord[top_key]  # Значение первого словаря
                bottom_key = list(bottom_dict.keys())[0]  # Ключ последнего словаря
                bottom_value = coord[bottom_key]  # Значение последнего словаря

                # top_value = utils.calculate_center_of_coordinates(top_value)
                # bottom_value = utils.calculate_center_of_coordinates(bottom_value)

                if swipe_direction_down:
                    self.terminal.swipe(*bottom_value)
                else:
                    self.terminal.swipe(*top_value)
                time.sleep(2)
                last_coord = self.get_text_coordinates(text=items)
                if last_coord == coord:
                    swipe_direction_down = not swipe_direction_down

            except IndexError as error:
                self.logger.error(f"IndexError {error}")
                if drop_down_locator is not None:
                    self.terminal.tap(drop_down_locator)
                continue
            except Exception as error:
                self.logger.error(f"Exception {error}")
                continue

    @helpers_decorators.retry
    def get_image_coordinates(self,
                              image: Union[bytes, np.ndarray, Image.Image, str],
                              full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                              threshold: Optional[float] = 0.7,
                              ) -> Union[Tuple[int, int, int, int], None]:
        """
        Находит координаты наиболее вероятного совпадения частичного изображения в полном изображении.

        Args:
            image (Union[bytes, np.ndarray, Image.Image, str]):
                Частичное изображение или путь к файлу, которое нужно найти внутри полного изображения.
            full_image (Union[bytes, np.ndarray, Image.Image, str], optional):
                Полное изображение или путь к файлу. По умолчанию None, в этом случае используется скриншот экрана.
            threshold (float, optional):
                Минимальный порог совпадения для считывания совпадения допустимым. По умолчанию 0.7.

        Usages:
            app.get_image_coordinates('path/to/partial_image.png', 'path/to/full_image.png')
            app.get_image_coordinates('path/to/partial_image.png', threshold=0.8)

        Returns:
            Union[Tuple[int, int, int, int], None]:
                Кортеж с координатами наиболее вероятного совпадения (x1, y1, x2, y2)
                или None, если совпадение не найдено.
        """
        if full_image is None:
            screenshot = self._get_screenshot_as_base64_decoded()
            big_image = self._to_ndarray(image=screenshot, grayscale=True)
        else:
            big_image = self._to_ndarray(image=full_image, grayscale=True)  # Загрузка полного изображения

        small_image = self._to_ndarray(image=image, grayscale=True)  # Загрузка частичного изображения

        # Сопоставление частичного изображения и снимка экрана
        max_val_loc = self._multi_scale_matching(full_image=big_image, template_image=small_image,
                                                 threshold=threshold)
        if max_val_loc is None:
            return None

        max_val, max_loc = max_val_loc

        if not max_val >= threshold:  # Если наибольшее значение совпадения не превышает порога, возвращаем None
            self.logger.error("find_coordinates_by_image(): Совпадений не найдено")
            return None

        # Вычисляем координаты левого верхнего и правого нижнего углов найденного совпадения
        left = int(max_loc[0])
        top = int(max_loc[1])
        width = small_image.shape[1]
        height = small_image.shape[0]
        right = left + width
        bottom = top + height

        return int(left), int(top), int(right), int(bottom)  # Возвращаем координаты наиболее вероятного совпадения

    @helpers_decorators.retry
    def get_inner_image_coordinates(self,
                                    outer_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                    inner_image_path: Union[bytes, np.ndarray, Image.Image, str],
                                    threshold: float = 0.9) -> Union[Tuple[int, int, int, int], None]:
        """
        Находит изображение на экране и внутри него находит другое изображение (внутреннее).

        Args:
            outer_image_path (Union[bytes, np.ndarray, Image.Image, str]):
                Внешнее изображение или путь к файлу, которое нужно найти на экране.
            inner_image_path (Union[bytes, np.ndarray, Image.Image, str]):
                Внутреннее изображение или путь к файлу, которое нужно найти внутри внешнего изображения.
            threshold (float, optional):
                Пороговое значение сходства для шаблонного сопоставления. По умолчанию 0.9.

        Usages:
            app.get_inner_image_coordinates('path/to/outer_image.png', 'path/to/inner_image.png')
            app.get_inner_image_coordinates('path/to/outer_image.png', 'path/to/inner_image.png', threshold=0.8)

        Returns:
            Union[Tuple[int, int, int, int], None]:
                Координаты внутреннего изображения относительно экрана в формате (x1, y1, x2, y2).
                Если внутреннее изображение не найдено, возвращает None.

        Note:
            Повторяет выполнение 3 раза при неудаче.
        """
        # Получаем разрешение экрана
        screen_width, screen_height = self.terminal.get_screen_resolution()

        # Захватываем скриншот
        screenshot = base64.b64decode(self.driver.get_screenshot_as_base64())

        # Читаем скриншот
        full_image = self._to_ndarray(image=screenshot, grayscale=True)

        # Прочитать внешнее изображение
        outer_image = self._to_ndarray(image=outer_image_path, grayscale=True)

        # Прочитать внутреннее изображение
        inner_image = self._to_ndarray(image=inner_image_path, grayscale=True)

        # Вычисляем коэффициенты масштабирования
        width_ratio = screen_width / full_image.shape[1]
        height_ratio = screen_height / full_image.shape[0]

        # ...
        inner_image = cv2.resize(inner_image, None, fx=width_ratio, fy=height_ratio)
        outer_image = cv2.resize(outer_image, None, fx=width_ratio, fy=height_ratio)

        outer_max_val, outer_max_loc = self._multi_scale_matching(full_image=full_image, template_image=outer_image,
                                                                  threshold=threshold)

        # Проверить, превышает ли максимальное значение сходства для внешнего изображения пороговое значение
        if outer_max_val >= threshold:
            # Получить размеры внешнего изображения
            outer_height, outer_width = outer_image.shape

            # Вычислить координаты внешнего изображения на экране
            outer_top_left = outer_max_loc
            outer_bottom_right = (outer_top_left[0] + outer_width, outer_top_left[1] + outer_height)

            # Извлечь область интереса (ROI), содержащую внешнее изображение
            outer_roi = full_image[outer_top_left[1]:outer_bottom_right[1], outer_top_left[0]:outer_bottom_right[0]]

            inner_max_val, inner_max_loc = self._multi_scale_matching(full_image=outer_roi, template_image=inner_image,
                                                                      threshold=threshold)

            # Проверить, превышает ли максимальное значение сходства для внутреннего изображения пороговое значение
            if inner_max_val >= threshold:
                # Получить размеры внутреннего изображения
                inner_height, inner_width = inner_image.shape

                # Вычислить координаты внутреннего изображения относительно экрана
                inner_top_left = (outer_top_left[0] + inner_max_loc[0], outer_top_left[1] + inner_max_loc[1])
                inner_bottom_right = (inner_top_left[0] + inner_width, inner_top_left[1] + inner_height)

                # Вернуть координаты внутреннего изображения относительно экрана
                return inner_top_left + inner_bottom_right

        # Вернуть None, если внутреннее изображение не найдено
        return None

    def _is_image_on_the_screen(self,
                                image: Union[bytes, np.ndarray, Image.Image, str],
                                threshold: float = 0.9) -> bool:
        """
        Сравнивает, присутствует ли заданное изображение на экране.

        Args:
            image (Union[bytes, np.ndarray, Image.Image, str]): Изображение для поиска на экране.
                Может быть в формате байтов, массива numpy, объекта Image.Image или строки с путем до файла.
            threshold (float): Пороговое значение схожести части изображения со снимком экрана.

        Returns:
            bool: Возвращает `True`, если изображение найдено на экране, иначе `False`.

        Raises:
            cv2.error: Ошибки, связанные с OpenCV.
            AssertionError: Ошибки, связанные с неверными размерами изображений.
            Exception: Остальные исключения.
        """
        try:
            screenshot = self._get_screenshot_as_base64_decoded()

            # Чтение снимка экрана и частичного изображения
            full_image = self._to_ndarray(image=screenshot, grayscale=True)
            small_image = self._to_ndarray(image=image, grayscale=True)

            # Проверка размеров изображений
            if small_image.shape[0] > full_image.shape[0] or small_image.shape[1] > full_image.shape[1]:
                self.logger.error("Частичное изображение больше снимка экрана.")
                return False

            # Сопоставление частичного изображения и снимка экрана
            max_val, max_loc = self._multi_scale_matching(full_image=full_image, template_image=small_image,
                                                          threshold=threshold)

            return max_val > threshold
        except cv2.error as e:
            self.logger.error(f"is_image_on_the_screen(): {e}")
            return False
        except AssertionError as e:
            self.logger.error(f"is_image_on_the_screen(): {e}")
            return False

    @staticmethod
    def _multi_scale_matching(full_image: np.ndarray,
                              template_image: np.ndarray,
                              threshold: float = 0.8,
                              return_raw: bool = False):
        origin_width, origin_height = template_image.shape[::-1]  # Исходный размер шаблона

        # Цикл по различным масштабам, включая масштабы больше 1.0 для "растягивания"
        for scale in np.concatenate([np.linspace(0.2, 1.0, 10)[::-1], np.linspace(1.1, 2.0, 10)]):

            # Изменение размера изображения и сохранение масштаба
            resized = cv2.resize(full_image, (int(full_image.shape[1] * scale), int(full_image.shape[0] * scale)))

            # Если измененный размер становится меньше шаблона, прерываем цикл
            if resized.shape[0] < origin_height or resized.shape[1] < origin_width:
                continue

            # Сопоставление шаблона
            result = cv2.matchTemplate(resized, template_image, cv2.TM_CCOEFF_NORMED)

            _, max_val, _, max_loc = cv2.minMaxLoc(result)

            if max_val > threshold:
                if return_raw:
                    return result
                # Преобразование координат обратно к оригинальному масштабу
                max_loc_original = (int(max_loc[0] / scale), int(max_loc[1] / scale))
                return max_val, max_loc_original

        if return_raw:
            return None
        return 0, (0, 0)

    def is_text_on_ocr_screen(self,
                              text: str,
                              screen: Union[bytes, np.ndarray, Image.Image, str] = None,
                              language: str = 'rus') -> bool:
        """
        Проверяет, присутствует ли заданный текст на экране.
        Распознавание текста производит с помощью библиотеки pytesseract.

        Аргументы:
        - text (str): Текст, который нужно найти на экране.
        - screen (bytes, optional): Скриншот в формате bytes. Если не указан, будет захвачен скриншот с помощью `self.driver`.
        - language (str): Язык распознавания текста. Значение по умолчанию: 'rus'.

        Возвращает:
        - bool: True, если заданный текст найден на экране. False в противном случае.
        """
        try:
            if screen is None:
                screenshot = self._get_screenshot_as_base64_decoded()
                image = self._to_ndarray(screenshot)
            else:
                image = self._to_ndarray(screen)

            # Адаптивная бинаризация
            threshold = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2)

            # Преобразование двоичного изображения в текст
            custom_config = r'--oem 3 --psm 6'

            ocr_text = pytesseract.image_to_string(threshold, lang=language, config=custom_config)

            # Проверка наличия заданного текста в распознанном тексте
            return text.lower() in ocr_text.lower()
        except cv2.error as e:
            self.logger.error(f"is_text_on_ocr_screen(): {e}")
            return False
        except pytesseract.TesseractError as e:
            self.logger.error(f"is_text_on_ocr_screen(): {e}")
            return False
        except AssertionError as e:
            self.logger.error(f"is_text_on_ocr_screen(): {e}")
            return False

    @helpers_decorators.retry
    def get_many_coordinates_of_image(self,
                                      image: Union[bytes, np.ndarray, Image.Image, str],
                                      full_image: Union[bytes, np.ndarray, Image.Image, str] = None,
                                      cv_threshold: float = 0.7,
                                      coord_threshold: int = 5) -> Union[List[Tuple], None]:
        """
        Находит все вхождения частичного изображения внутри полного изображения.

        Args:
            image (Union[bytes, np.ndarray, Image.Image, str]):
                Частичное изображение или путь к файлу, которое нужно найти внутри полного изображения.
            full_image (Union[bytes, np.ndarray, Image.Image, str], optional):
                Полное изображение или путь к файлу. По умолчанию None, в этом случае используется скриншот экрана.
            cv_threshold (float, optional):
                Минимальный порог совпадения для считывания совпадения допустимым. По умолчанию 0.7.
            coord_threshold (int, optional):
                Максимальное различие между значениями x и y двух кортежей, чтобы они считались слишком близкими друг к другу.
                По умолчанию 5 пикселей.

        Usages:
            app.et_many_coordinates_of_image('path/to/partial_image.png', 'path/to/full_image.png')
            app.get_many_coordinates_of_image('path/to/partial_image.png', cv_threshold=0.8, coord_threshold=10)

        Returns:
            Union[List[Tuple], None]:
                Список кортежей, содержащий расположение каждого найденного совпадения в формате (x1, y1, x2, y2).
                Если совпадений не найдено, возвращает None.

        Note:
            При неудаче повторяет выполнение, до трёх раз.
        """

        if full_image is None:
            screenshot = self._get_screenshot_as_base64_decoded()
            big_image = self._to_ndarray(image=screenshot, grayscale=True)
        else:
            big_image = self._to_ndarray(image=full_image, grayscale=True)  # Загрузка полного изображения

        small_image = self._to_ndarray(image=image, grayscale=True)  # Загрузка частичного изображения

        result = self._multi_scale_matching(full_image=big_image, template_image=small_image,
                                            return_raw=True, threshold=cv_threshold)

        # Получить все совпадения выше порога
        locations = np.where(result >= cv_threshold)  # Нахождение всех совпадений выше порога
        matches = list(zip(*locations[::-1]))  # Преобразование координат в список кортежей

        # Фильтрация слишком близких совпадений
        unique_list = []  # Создаем пустой список для хранения уникальных кортежей
        for (x1_coordinate, y1_coordinate) in matches:  # Итерируемся по списку кортежей
            exclude = False  # Инициализируем флаг exclude значением False
            for (x2_coordinate, y2_coordinate) in unique_list:  # Итерируемся по уникальным кортежам
                if abs(x1_coordinate - x2_coordinate) <= coord_threshold and abs(
                        y1_coordinate - y2_coordinate) <= coord_threshold:
                    # Если различие между значениями x и y двух кортежей меньше или равно порогу,
                    # помечаем exclude как True и выходим из цикла
                    exclude = True
                    break
            if not exclude:  # Если exclude равно False, добавляем кортеж в unique_list
                unique_list.append((x1_coordinate, y1_coordinate))
        matches = unique_list

        if not matches:
            self.logger.error(f"_find_many_coordinates_by_image() NO MATCHES, {image=}")
            return None

        # Добавляем правый нижний угол к каждому найденному совпадению
        matches_with_corners = []
        for match in matches:
            x_coordinate, y_coordinate = match
            width, height = small_image.shape[::-1]
            top_left = (x_coordinate, y_coordinate)
            bottom_right = (x_coordinate + width, y_coordinate + height)
            matches_with_corners.append((top_left + bottom_right))

        return matches_with_corners

    @helpers_decorators.retry
    def get_text_coordinates(self,
                             text: Union[str, List[str]],
                             image: Union[bytes, str, Image.Image, np.ndarray] = None,
                             language: str = 'rus'
                             ) -> Union[tuple[int, ...], Dict[str, tuple], None]:
        """
        Возвращает координаты области с указанным текстом на предоставленном изображении или снимке экрана.

        Args:
        - text (str): Искомый текст.
        - image (bytes, str, Image.Image, np.ndarray, опционально): Изображение, на котором осуществляется поиск текста.
          Если не указано, будет использован снимок экрана. По умолчанию None.
        - language (str, опционально): Язык для распознавания текста. По умолчанию 'rus'.

        Usages:
            app.get_text_coordinates("Hello, world!")
            app.get_text_coordinates("Привет, мир!", language='rus')
            app.get_text_coordinates("Hello, world!", image='path/to/image.png')

        Returns:
        - Union[Tuple[int, int, int, int], None]: Координаты области с текстом или None, если текст не найден.
        """
        if not image:
            # Получаем снимок экрана, если изображение не предоставлено
            screenshot = self._get_screenshot_as_base64_decoded()  # Получение снимка экрана в формате base64
            image = self._to_ndarray(image=screenshot,
                                     grayscale=True)  # Преобразование снимка экрана в массив numpy и преобразование в оттенки серого
        else:
            # Если предоставлено, то преобразуем
            image = self._to_ndarray(image=image,
                                     grayscale=True)  # Преобразование изображения в массив numpy и преобразование в оттенки серого
        image = cv2.medianBlur(image, 3)  # + устранение шума
        image = cv2.convertScaleAbs(image, alpha=1.5, beta=30)

        # Адаптивная бинаризация
        threshold = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2)
        oem_config = r'--oem 3 --psm 6'
        # Выполнение OCR с помощью PyTesseract
        data = pytesseract.image_to_data(threshold,
                                         lang=language,
                                         output_type=pytesseract.Output.DICT,
                                         config=oem_config)  # Использование PyTesseract для распознавания текста и получения информации о распознанных словах

        formatted_data = {}

        for i in range(len(data['text'])):
            word_text = data['text'][i]  # Текст слова
            left = int(data['left'][i])  # Координата левой границы слова
            top = int(data['top'][i])  # Координата верхней границы слова
            width = int(data['width'][i])  # Ширина слова
            height = int(data['height'][i])  # Высота слова
            coordinates = [left, top, left + width, top + height]  # Координаты рамки слова

            if word_text:
                if i not in formatted_data:
                    formatted_data[i] = {}
                formatted_data[i] = {'text': word_text,
                                     'coordinates': coordinates}  # Сохранение информации о слове и его координатах
        # Инициализировать переменные для последовательности слов и соответствующих координат
        current_sequence = []  # Текущая последовательность слов
        result_coordinates = []  # Координаты текущей последовательности слов
        result_dict = {}
        flag_return_list = False

        if isinstance(text, str):
            text = [text]
            flag_return_list = True
        for element in text:
            # Разбить искомый текст на отдельные слова
            words = element.lower().split(' ')  # Разделение искомого текста на отдельные слова
            for word_data in formatted_data.values():
                word = word_data['text'].lower()  # Текущее слово
                coordinates = word_data['coordinates']  # Координаты слова

                if word in words:
                    current_sequence.append(word)  # Добавление слова в текущую последовательность
                    result_coordinates.append(coordinates)  # Добавление координат слова в результат

            if utils.is_list_in_list(small_list=words, big_list=current_sequence):
                # Обрезка списка координат до размерности заданной последовательности слов
                if len(current_sequence) > len(words):
                    result_coordinates = result_coordinates[-len(words):]

                # Если найдена последовательность слов, вернуть соответствующие координаты
                top_left = tuple(map(int, result_coordinates[0][:2]))  # Верхний левый угол рамки
                bottom_right = tuple(map(int, result_coordinates[-1][2:]))  # Нижний правый угол рамки
                if flag_return_list:
                    return top_left + bottom_right
                words = ' '.join(words)
                result_dict[words] = top_left + bottom_right
            current_sequence = []  # Сброс текущей последовательности слов
            result_coordinates = []  # Сброс координат последовательности слов
        return result_dict

    def draw_by_coordinates(self,
                            image: Union[bytes, str, Image.Image, np.ndarray] = None,
                            coordinates: Tuple[int, int, int, int] = None,
                            top_left: Tuple[int, int] = None,
                            bottom_right: Tuple[int, int] = None,
                            path: str = None) -> bool:
        """
        Рисует прямоугольник на предоставленном изображении или снимке экрана с помощью драйвера.

        Args:
            image (Union[bytes, str, Image.Image, np.ndarray], optional): Изображение для рисования. По умолчанию None.
            coordinates (Tuple[int, int, int, int], optional): Координаты прямоугольника (x1, y1, x2, y2). По умолчанию None.
            top_left (Tuple[int, int], optional): Верхняя левая точка прямоугольника. По умолчанию None.
            bottom_right (Tuple[int, int], optional): Нижняя правая точка прямоугольника. По умолчанию None.
            path (str, optional): Путь для сохранения изображения. По умолчанию None.

        Usages:
            draw_by_coordinates(image=image_bytes, coordinates=(10, 20, 30, 40), path='path/to/save/image.png')
            draw_by_coordinates(top_left=(10, 20), bottom_right=(30, 40))

        Returns:
            bool: True, если операция выполнена успешно, иначе False.

        Raises:
            WebDriverException: Если возникают проблемы с WebDriver.
            cv2.error: Если возникают проблемы с OpenCV.

        Notes:
            - Если изображение не предоставлено, будет использован текущий снимок экрана.
            - Если не указаны верхняя левая и нижняя правая точки, будут использованы координаты.
        """
        try:
            if image is None:
                # Если изображение не предоставлено, получаем снимок экрана с помощью драйвера
                screenshot = self._get_screenshot_as_base64_decoded()
                image = self._to_ndarray(screenshot)
            else:
                image = self._to_ndarray(image)

            # Если верхняя левая и нижняя правая точки не предоставлены, используем координаты для определения
            # прямоугольника
            if not top_left and not bottom_right:
                top_left = (coordinates[0], coordinates[1])
                bottom_right = (coordinates[2], coordinates[3])

            # Сохраняем снимок экрана с нарисованным прямоугольником
            if path is None:
                path = "screenshot_with_text_coordinates.png"
            path = os.path.join(path)
            cv2.rectangle(image, top_left, bottom_right, (0, 255, 0), 2)
            cv2.imwrite(path, image)

            return True
        except cv2.error as e:
            # Обработка исключения cv2.error
            self.logger.error(f'draw_by_coordinates() cv2.error: {e}')
            return False

    @staticmethod
    def is_rgb(image: np.ndarray) -> bool:
        """
        Проверяет, является ли изображение цветным (RGB).

        Аргументы:
        - image: np.ndarray - Входное изображение в формате NumPy ndarray.

        Возвращает:
        - bool - True, если изображение является цветным (RGB), False - в противном случае.
        """
        return len(image.shape) == 3 and image.shape[2] == 3 or image.ndim == 3 or image.ndim == '3'

    @staticmethod
    def is_grayscale(image: np.ndarray) -> bool:
        """
        Проверяет, является ли изображение оттенков серого.

        Аргументы:
        - image: np.ndarray - Входное изображение в формате NumPy ndarray.

        Возвращает:
        - bool - True, если изображение является оттенков серого, False - в противном случае.
        """
        return len(image.shape) == 2 or (
                len(image.shape) == 3 and image.shape[2] == 1) or image.ndim == 2 or image.ndim == '2'

    def to_grayscale(self, image: np.ndarray) -> np.ndarray:
        """
        Преобразует изображение в оттенки серого.

        Аргументы:
        - image: np.ndarray - Входное изображение в формате ndarray.

        Возвращает:
        - np.ndarray - Преобразованное изображение в оттенках серого.
        """
        # Проверяем, является ли изображение в формате RGB
        if self.is_rgb(image):
            # Если да, то преобразуем его в оттенки серого
            gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            # Приводим значения пикселей к диапазону от 0 до 255
            gray_image = cv2.convertScaleAbs(gray_image)
            return gray_image
        # Иначе, возвращаем изображение без изменений
        return image

    def _to_ndarray(self, image: Union[bytes, np.ndarray, Image.Image, str], grayscale: bool = True) -> np.ndarray:
        """
        Преобразует входные данные из различных типов в ndarray (NumPy array).

        Аргументы:
        - image: Union[bytes, np.ndarray, Image.Image, str] - Входные данные,
          представляющие изображение. Может быть типами bytes, np.ndarray, PIL Image или str.

        Возвращает:
        - np.ndarray - Преобразованный массив NumPy (ndarray) представляющий изображение.
        """
        # Если входные данные являются массивом байтов, преобразовать их в массив NumPy
        if isinstance(image, bytes):
            image = cv2.imdecode(np.frombuffer(image, np.uint8), cv2.IMREAD_COLOR)

        # Если входные данные являются строкой с путем к файлу, открыть изображение и преобразовать в массив NumPy
        if isinstance(image, str):
            # image = np.array(Image.open(image))
            image = cv2.imread(image, cv2.IMREAD_COLOR)

        # Если входные данные являются объектом PIL Image, преобразовать его в массив NumPy
        if isinstance(image, Image.Image):
            image = np.array(image)

        # Вернуть преобразованный массив NumPy
        if grayscale:
            return self.to_grayscale(image=image)
        return image

    def _save_screenshot(self, path: str = '', filename: str = 'screenshot.png') -> bool:
        """
        Сохраняет скриншот экрана в указанный файл.

        Args:
            path (str, optional): Путь к директории, где будет сохранен скриншот. По умолчанию пустая строка, что означает текущую директорию.
            filename (str, optional): Имя файла, в который будет сохранен скриншот. По умолчанию 'screenshot.png'.

        Usages:
            save_screenshot(path='/path/to/save', filename='my_screenshot.png')
            save_screenshot(filename='another_screenshot.png')
            save_screenshot()

        Returns:
            bool: True, если скриншот успешно сохранен, иначе False.

        Raises:
            Exception: В случае, если возникают проблемы при сохранении скриншота.

        Notes:
            - Если путь не указан, скриншот будет сохранен в текущей директории.
            - Если имя файла не указано, будет использовано имя 'screenshot.png'.
        """
        screenshot = self._get_screenshot_as_base64_decoded()
        path_to_file = os.path.join(path, filename)
        with open(path_to_file, "wb") as f:
            f.write(screenshot)
        return True

    def show_screen(self):
        """
        Выводит на экран теста скриншот текущего экрана устройства.
        Код не будет продолжать выполнятся, пока изображение не закрыть.
        Метод для отладки.
        """
        cv2.imshow('screen', self._to_ndarray(self._get_screenshot_as_base64_decoded()))
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    def _get_screenshot_as_base64_decoded(self) -> bytes:
        """
        Получает скриншот экрана, кодирует его в формате Base64, а затем декодирует в байты.

        Args:
            Метод не принимает аргументов.

        Usages:
            screenshot_bytes = self._get_screenshot_as_base64_decoded()

        Returns:
            bytes: Декодированные байты скриншота, обычно в формате PNG.

        Raises:
            WebDriverException: Если не удается получить скриншот.

        Notes:
            - Этот метод предназначен для внутреннего использования и может быть вызван другими методами класса.
            - Скриншот возвращается в формате PNG.
            - Исходный скриншот получается в формате Base64, который затем кодируется в UTF-8 и декодируется обратно в байты.
        """
        screenshot = self.driver.get_screenshot_as_base64().encode('utf-8')
        screenshot = base64.b64decode(screenshot)
        return screenshot



import inspect
import logging
import base64
import os
import re
import subprocess
import sys
import time
import traceback
from typing import Dict, Any, Union, Tuple

from selenium.common import NoSuchDriverException, InvalidSessionIdException

# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath( __file__))))  # The sys.path.append line adds the
# parent directory of the tests directory to the Python module search path, allowing you to import modules from the
# root folder.
from appium_extended_helpers.helpers_decorators import log_debug


class Terminal:
    base = None
    transport = None
    driver = None
    logger = None

    def __init__(self, base, logger: logging.Logger = None, log_level: int = logging.INFO, log_path: str = ''):
        self.base = base
        self.transport = base.transport
        self.driver = base.driver
        self.logger = base.logger
        if logger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(log_level)
        if bool(log_path):
            if not log_path.endswith('.log'):
                log_path = log_path + '.log'
            file_handler = logging.FileHandler(log_path)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

    def __del__(self):
        if self.transport is not None:
            self.transport.ssh.close()

    @log_debug()
    def adb_shell(self, command: str, args: str = "", tries: int = 3) -> Any:
        for _ in range(tries):
            try:
                return self.driver.execute_script("mobile: shell", {'command': command, 'args': [args]})
            except NoSuchDriverException:
                self.base.reconnect()
            except InvalidSessionIdException:
                self.base.reconnect()
            except KeyError as e:
                self.logger.error(e)
                traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
                self.logger.error(traceback_info)

    @log_debug()
    def push(self, source_path: str, remote_server_path: str, filename: str, destination: str, udid: str) -> bool:
        """
        Копирует файл или директорию на подключенное устройство через Appium сервер.

        Аргументы:
            driver: Appium WebDriver объект.
            source (str): Путь к копируемому файлу или директории на локальной машине.
            destination (str): Путь назначения на устройстве.

        Возвращает:
            bool: True, если файл или директория были успешно скопированы, False в противном случае.
        """
        if self.transport is None:
            raise AssertionError(
                f"Метод {inspect.currentframe().f_code.co_name} ssh_username, ssh_password в методе connect()")
        try:
            source_file_path = os.path.join(source_path, filename)
            remote_file_path = os.path.join(remote_server_path, filename)
            destination_file_path = f"{destination}/{filename}"
            self.transport.scp.put(files=source_file_path, remote_path=remote_file_path)
            stdin, stdout, stderr = self.transport.ssh.exec_command(
                f'adb -s {udid} push {remote_file_path} {destination_file_path}')
            stdout_exit_status = stdout.channel.recv_exit_status()
            lines = stdout.readlines()
            output = ''.join(lines)
            if stdout_exit_status != 0:
                self.logger.error(f"{inspect.currentframe().f_code.co_name} {output=}")
                return False
            self.logger.debug(f"{inspect.currentframe().f_code.co_name} {output=}")
            return True
        except NoSuchDriverException:
            self.base.reconnect()
        except InvalidSessionIdException:
            self.base.reconnect()
        except IOError as e:
            self.logger.error("appium_extended_terminal.push()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def pull(self, source: str, destination: str) -> bool:
        """Извлекает файл с устройства по пути `source` и сохраняет его по пути `destination` на компьютере.

        Аргументы:
            source (str): Путь к файлу на устройстве.
            destination (str): Путь, по которому файл должен быть сохранен на компьютере.

        Возвращает:
            bool: True, если файл успешно извлечен и сохранен, False в противном случае.
        """
        try:
            file_contents_base64 = self.driver.assert_extension_exists('mobile: pullFile'). \
                execute_script('mobile: pullFile', {'remotePath': source})
            if not file_contents_base64:
                return False
            decoded_contents = base64.b64decode(file_contents_base64)
            with open(destination, 'wb') as file:
                file.write(decoded_contents)
            return True
        except NoSuchDriverException:
            self.base.reconnect()
        except InvalidSessionIdException:
            self.base.reconnect()
        except IOError as e:
            self.logger.error("appium_extended_terminal.pull")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def start_activity(self, package: str, activity: str) -> bool:
        """
        Запускает активити на подключенном устройстве.

        Аргументы:
            package (str): Название пакета.
            activity (str): Название запускаемой активити.

        Возвращает:
            bool: True, если активность была успешно запущена, False в противном случае.
        """
        try:
            self.adb_shell(command="am", args=f"start -n {package}/{activity}")
            return True
        except KeyError as e:
            self.logger.error("appium_extended_terminal.start_activity()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def get_current_app_package(self) -> Union[str, None]:
        """
        Получает пакет текущего запущенного приложения на устройстве с помощью ADB.

        Возвращает:
            str: Название пакета текущего запущенного приложения, либо None, если произошла ошибка.
        """
        try:
            result = self.adb_shell(command="dumpsys", args="window windows")
            lines = result.split('\n')
            for line in lines:
                if 'mCurrentFocus' in line or 'mFocusedApp' in line:
                    matches = re.search(r'(([A-Za-z]{1}[A-Za-z\d_]*\.)+([A-Za-z][A-Za-z\d_]*)/)', line)
                    if matches:
                        return matches.group(1)[:-1]  # removing trailing slash
            return None
        except KeyError as e:
            # Логируем ошибку, если возникло исключение
            self.logger.error("appium_extended_terminal.get_current_app_package()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return None

    @log_debug()
    def close_app(self, package: str) -> bool:
        """
        Принудительно останавливает указанный пакет с помощью ADB.

        Аргументы:
            package (str): Название пакета приложения для закрытия.

        Возвращает:
            bool: True, если приложение успешно закрыто, False в противном случае.
        """
        try:
            self.adb_shell(command="am", args=f"force-stop {package}")
            return True
        except KeyError as e:
            self.logger.error("appium_extended_terminal.close_app()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def reboot_app(self, package: str, activity: str) -> bool:
        """
        Перезапускает приложение, закрывая его и затем запуская указанную активность.

        Аргументы:
            package (str): Название пакета приложения.
            activity (str): Название активности для запуска.

        Возвращает:
            bool: True, если перезапуск приложения выполнен успешно, False в противном случае.
        """
        # Закрытие приложения
        if not self.close_app(package=package):
            return False

        # Запуск указанной активности
        if not self.start_activity(package=package, activity=activity):
            return False

        return True

    @log_debug()
    def install_app(self, source: str, remote_server_path: str, filename: str, udid: str) -> bool:
        """
        Устанавливает указанный пакет с помощью Appium.
        Дублирует команду драйвера. Добавлено для интуитивности.

        Аргументы:
            package (str): Название пакета приложения для установки.

        Возвращает:
            bool: True, если приложение успешно удалено, False в противном случае.
        """
        if self.transport is None:
            raise AssertionError(
                f"Метод {inspect.currentframe().f_code.co_name} нельзя использовать, если не переданы server, port, username, password для обеспечения ssh соединения")
        try:
            source_filepath = os.path.join(source, filename)
            destination_filepath = os.path.join(remote_server_path, filename)
            self.transport.scp.put(files=source_filepath, remote_path=destination_filepath)
            stdin, stdout, stderr = self.transport.ssh.exec_command(
                f'adb -s {udid} install -r {destination_filepath}')
            stdout_exit_status = stdout.channel.recv_exit_status()
            lines = stdout.readlines()
            output = ''.join(lines)
            if stdout_exit_status != 0:
                self.logger.error(f"{inspect.currentframe().f_code.co_name} {output=}")
                return False
            self.logger.debug(f"{inspect.currentframe().f_code.co_name} {output=}")
            return True
        except IOError as e:
            self.logger.error("appium_extended_terminal.push()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def is_app_installed(self, package) -> bool:
        """
        Проверяет, установлен ли пакет.
        """
        self.logger.debug(f"is_app_installed() < {package=}")

        try:
            result = self.adb_shell(command="pm", args="list packages")
            # Фильтруем пакеты
            if any([line.strip().endswith(package) for line in result.splitlines()]):
                self.logger.debug("is_app_installed() > True")
                return True
            self.logger.debug("is_app_installed() > False")
            return False
        except KeyError as e:
            self.logger.error("appium_extended_terminal.is_app_installed() > False")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def uninstall_app(self, package: str) -> bool:
        """
        Удаляет указанный пакет с помощью ADB.

        Аргументы:
            package (str): Название пакета приложения для удаления.

        Возвращает:
            bool: True, если приложение успешно удалено, False в противном случае.
        """
        try:
            self.driver.remove_app(app_id=package)
            return True
        except NoSuchDriverException:
            self.base.reconnect()
        except InvalidSessionIdException:
            self.base.reconnect()
        except KeyError as e:
            self.logger.error("appium_extended_terminal.uninstall_app()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def press_home(self) -> bool:
        """
        Отправляет событие нажатия кнопки Home на устройство с помощью ADB.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.input_keycode(keycode="KEYCODE_HOME")
            return True
        except KeyError as e:
            self.logger.error("appium_extended_terminal.press_home()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def press_back(self) -> bool:
        """
        Отправляет событие нажатия кнопки Back на устройство с помощью ADB.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.input_keycode(keycode="KEYCODE_BACK")
            return True
        except KeyError as e:
            self.logger.error("appium_extended_terminal.press_back()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def press_menu(self) -> bool:
        """
        Отправляет событие нажатия кнопки Menu на устройство с помощью ADB.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.input_keycode(keycode="KEYCODE_MENU")
            return True
        except KeyError as e:
            self.logger.error("appium_extended_terminal.press_menu()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def input_keycode_num_(self, num: int) -> bool:
        """
        Отправляет событие нажатия клавиши с числовым значением на устройство с помощью ADB.
        Допустимые значения: 0-9, ADD, COMMA, DIVIDE, DOT, ENTER, EQUALS

        Аргументы:
            num (int): Числовое значение клавиши для нажатия.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.adb_shell(command="input", args=f"keyevent KEYCODE_NUMPAD_{num}")
            return True
        except KeyError as e:
            self.logger.error("appium_extended_terminal.input_keycode_num_()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def input_keycode(self, keycode: str) -> bool:
        """
        Вводит указанный код клавиши на устройстве с помощью ADB.

        Аргументы:
            keycode (str): Код клавиши для ввода.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.adb_shell(command="input", args=f"keyevent {keycode}")
            return True
        except KeyError as e:
            self.logger.error("appium_extended_terminal.input_keycode()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def input_by_virtual_keyboard(self, key: str, keyboard: Dict[str, tuple]) -> bool:
        """
        Вводит строку символов с помощью виртуальной клавиатуры.

        Аргументы:
            key (str): Строка символов для ввода.
            keyboard (dict): Словарь с маппингом символов на координаты нажатий.

        Возвращает:
            bool: True, если ввод выполнен успешно, False в противном случае.
        """
        try:
            for char in key:
                # Вызываем функцию tap с координатами, соответствующими символу char
                self.tap(x=keyboard[str(char)][0], y=keyboard[str(char)][1])
            return True
        except KeyError as e:
            # Логируем ошибку и возвращаем False в случае возникновения исключения
            self.logger.error("appium_extended_terminal.input_by_virtual_keyboard")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def input_text(self, text: str) -> bool:
        """
        Вводит указанный текст на устройстве с помощью ADB.

        Аргументы:
            text (str): Текст для ввода.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.adb_shell(command="input", args=f"text {text}")
            return True
        except KeyError as e:
            # Логируем ошибку, если возникло исключение
            self.logger.error("appium_extended_terminal.input_text()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def tap(self, x: int, y: int) -> bool:
        """
        Выполняет нажатие на указанные координаты на устройстве с помощью ADB.

        Аргументы:
            x: Координата X для нажатия.
            y: Координата Y для нажатия.

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.adb_shell(command="input", args=f"tap {str(x)} {str(y)}")
            return True
        except KeyError as e:
            # Логируем ошибку, если возникло исключение
            self.logger.error("appium_extended_terminal.tap()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def swipe(self, start_x: Union[str, int], start_y: Union[str, int],
              end_x: Union[str, int], end_y: Union[str, int], duration: int = 300) -> bool:
        """
        Выполняет свайп (перетаскивание) с одной точки на экране в другую на устройстве с помощью ADB.

        Аргументы:
            start_x: Координата X начальной точки свайпа.
            start_y: Координата Y начальной точки свайпа.
            end_x: Координата X конечной точки свайпа.
            end_y: Координата Y конечной точки свайпа.
            duration (int): Длительность свайпа в миллисекундах (по умолчанию 300).

        Возвращает:
            bool: True, если команда была успешно выполнена, False в противном случае.
        """
        try:
            self.adb_shell(command="input",
                           args=f"swipe {str(start_x)} {str(start_y)} {str(end_x)} {str(end_y)} {str(duration)}")
            return True
        except KeyError as e:
            # Логируем ошибку, если возникло исключение
            self.logger.error("appium_extended_terminal.swipe()")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def swipe_right_to_left(self, duration: int = 300) -> bool:
        window_size = self.get_screen_resolution()
        width = window_size[0]
        height = window_size[1]
        left = int(width * 0.1)
        right = int(width * 0.9)
        return self.swipe(start_x=right,
                          start_y=height // 2,
                          end_x=left,
                          end_y=height // 2,
                          duration=duration)

    @log_debug()
    def swipe_left_to_right(self, duration: int = 300) -> bool:
        window_size = self.get_screen_resolution()
        width = window_size[0]
        height = window_size[1]
        left = int(width * 0.1)
        right = int(width * 0.9)
        return self.swipe(start_x=left,
                          start_y=height // 2,
                          end_x=right,
                          end_y=height // 2,
                          duration=duration)

    @log_debug()
    def swipe_top_to_bottom(self, duration: int = 300) -> bool:
        window_size = self.get_screen_resolution()
        height = window_size[1]
        top = int(height * 0.1)
        bottom = int(height * 0.9)
        return self.swipe(start_x=top,
                          start_y=height // 2,
                          end_x=bottom,
                          end_y=height // 2,
                          duration=duration)

    @log_debug()
    def swipe_bottom_to_top(self, duration: int = 300) -> bool:
        window_size = self.get_screen_resolution()
        height = window_size[1]
        top = int(height * 0.1)
        bottom = int(height * 0.9)
        return self.swipe(start_x=bottom,
                          start_y=height // 2,
                          end_x=top,
                          end_y=height // 2,
                          duration=duration)

    @log_debug()
    def check_vpn(self, ip_address: str = '') -> bool:
        """
        Проверяет, активно ли VPN-соединение на устройстве с помощью ADB.

        Аргументы:
            ip (str): IP-адрес для проверки VPN-соединения. Если не указан, используется значение из конфигурации.

        Возвращает:
            bool: True, если VPN-соединение активно, False в противном случае.
        """
        try:
            output = self.adb_shell(command="netstat", args="")
            lines = output.split('\n')
            for line in lines:
                if ip_address in line and "ESTABLISHED" in line:
                    self.logger.debug("check_VPN() True")
                    return True
            self.logger.debug("check_VPN() False")
            return False
        except KeyError as e:
            # Логируем ошибку, если возникло исключение
            self.logger.error("appium_extended_terminal.check_VPN")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def stop_logcat(self) -> bool:
        """
        Останавливает выполнение logcat на устройстве с помощью ADB.

        Возвращает:
            bool: True, если выполнение logcat остановлено успешно, False в противном случае.
        """
        # Получаем список выполняющихся процессов logcat
        try:
            process_list = self.adb_shell(command="ps", args="")
        except KeyError as e:
            self.logger.error("appium_extended_terminal.stop_logcat")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        # Проходим по списку процессов и отправляем каждому сигнал SIGINT
        for process in process_list.splitlines():
            if "logcat" in process:
                pid = process.split()[1]
                try:
                    self.adb_shell(command="kill", args=f"-SIGINT {str(pid)}")
                except KeyError as e:
                    self.logger.error("appium_extended_terminal.stop_logcat")
                    self.logger.error(e)
                    traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
                    self.logger.error(traceback_info)
                    return False
        return True

    @log_debug()
    def know_pid(self, name: str) -> Union[int, None]:
        """
        Находит Process ID (PID) процесса по его имени, используя adb shell ps.

        Параметры:
            name (str): Имя процесса, PID которого нужно найти.

        Возвращает:
            Union[int, None]: PID процесса, если он найден, или None, если процесс не найден.
        """
        # Получение списка всех процессов с помощью adb shell ps
        processes = self.adb_shell(command="ps")
        if name not in processes:
            self.logger.error("know_pid() [Процесс не обнаружен]")
            return None
        # Разделение вывода на строки и удаление пустых строк
        lines = processes.strip().split('\n')
        # Проход по каждой строке вывода, начиная с 2-й строки, игнорируя заголовки
        for line in lines[1:]:
            # Разделение строки на столбцы по пробелам
            columns = line.split()
            # Проверка, что строка имеет не менее 9 столбцов
            if len(columns) >= 9:
                # Извлечение PID и имени процесса из соответствующих столбцов
                pid, process_name = columns[1], columns[8]
                # Сравнение имени процесса с искомым именем
                if name == process_name:
                    self.logger.debug(f"know_pid() > {str(pid)}")
                    return int(pid)
        self.logger.error("know_pid() [Процесс не обнаружен]")
        # Возврат None, если процесс с заданным именем не найден
        return None

    @log_debug()
    def is_process_exist(self, name) -> bool:
        """
        Проверяет, запущен ли процесс, используя adb shell ps.

        Параметры:
            name (str): Имя процесса.

        Возвращает:
            bool: True если процесс с указанным именем существует, False в ином случае.
        """
        # Получение списка всех процессов с помощью adb shell ps
        processes = self.adb_shell(command="ps")
        if name not in processes:
            self.logger.debug("is_process_exist() > False")
            return False
        # Разделение вывода на строки и удаление пустых строк
        lines = processes.strip().split('\n')
        # Проход по каждой строке вывода, начиная с 2-й строки, игнорируя заголовки
        for line in lines[1:]:
            # Разделение строки на столбцы по пробелам
            columns = line.split()
            # Проверка, что строка имеет не менее 9 столбцов
            if len(columns) >= 9:
                # Извлечение PID и имени процесса из соответствующих столбцов
                _, process_name = columns[1], columns[8]
                # Сравнение имени процесса с искомым именем
                if name == process_name:
                    self.logger.debug("is_process_exist() > True")
                    return True
        self.logger.debug("is_process_exist() > False")
        # Возврат None, если процесс с заданным именем не найден
        return False

    @log_debug()
    def run_background_process(self, command: str, args: str = "", process: str = "") -> bool:
        """
        Запускает процесс в фоновом режиме на устройстве Android.

        Аргументы:
            command (str): Команда для выполнения на устройстве.
            process (str): Название процесса, который будет запущен. По умолчанию "".
            Если process == "", то не будет проверяться его запуск в системе.

        Возвращает:
            bool: True, если процесс был успешно запущен, False в противном случае.
        """
        self.logger.debug(f"run_background_process() < {command=}")

        try:
            self.adb_shell(command=command, args=args + " nohup > /dev/null 2>&1 &")
            if process != "":
                time.sleep(1)
                if not self.is_process_exist(name=process):
                    return False
            return True
        except KeyError as e:
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False

    @log_debug()
    def kill_by_pid(self, pid: int) -> bool:
        """
        Отправляет сигнал SIGINT для остановки процесса по указанному идентификатору PID с помощью ADB.

        Аргументы:
            pid (str): Идентификатор PID процесса для остановки.

        Возвращает:
            bool: True, если процесс успешно остановлен, False в противном случае.
        """
        try:
            self.adb_shell(command="kill", args=f"-s SIGINT {str(pid)}")
        except KeyError as e:
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        return True

    @log_debug()
    def kill_by_name(self, name: str) -> bool:
        """
        Останавливает все процессы с указанным именем на устройстве с помощью ADB.

        Аргументы:
            name (str): Имя процесса для остановки.

        Возвращает:
            bool: True, если все процессы успешно остановлены, False в противном случае.
        """
        self.logger.debug(f"kill_by_name() < {name=}")
        try:
            self.adb_shell(command="pkill", args=f"-l SIGINT {str(name)}")
        except KeyError as e:
            self.logger.error("kill_by_name() > False")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        self.logger.debug("kill_by_name() > True")
        return True

    @log_debug()
    def kill_all(self, name: str) -> bool:
        """
        Останавливает все процессы, соответствующие указанному имени, на устройстве с помощью ADB.

        Аргументы:
            name (str): Имя процесса или шаблон имени для остановки.

        Возвращает:
            bool: True, если все процессы успешно остановлены, False в противном случае.
        """
        try:
            self.adb_shell(command="pkill", args=f"-f {str(name)}")
        except KeyError as e:
            self.logger.error("appium_extended_terminal.kill_all")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        return True

    @log_debug()
    def delete_files_from_internal_storage(self, path) -> bool:
        """
        Удаляет файлы из внутреннего хранилища устройства с помощью ADB.

        Аргументы:
            path (str): Путь к папке с файлами для удаления.

        Возвращает:
            bool: True, если файлы успешно удалены, False в противном случае.
        """
        try:
            self.adb_shell(command="rm", args=f"-rf {path}*")
        except KeyError as e:
            self.logger.error("appium_extended_terminal.delete_files_from_internal_storage")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        return True

    @log_debug()
    def delete_file_from_internal_storage(self, path: str, filename: str) -> bool:
        """
        Удаляет файл из внутреннего хранилища устройства с помощью ADB.

        Аргументы:
            path (str): Путь к папке с файлами для удаления.
            filename (str): Наименование файла.

        Возвращает:
            bool: True, если файл успешно удален, False в противном случае.
        """
        try:
            if path.endswith('/'):
                path = path[:-1]
            self.adb_shell(command="rm", args=f"-rf {path}/{filename}")
        except KeyError as e:
            self.logger.error("appium_extended_terminal.delete_file_from_internal_storage")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        return True

    @log_debug()
    def record_video(self, **options: Any) -> bool:
        """
        Начинает запись видео. 3 минуты максимум.

        Аргументы:
            filename (str): Имя файла для сохранения видео.

        Возвращает:
            bool: True, если запись видео успешно начата, False в противном случае.
        """
        try:
            self.driver.start_recording_screen(**options)
        except NoSuchDriverException:
            self.base.reconnect()
        except InvalidSessionIdException:
            self.base.reconnect()
        except KeyError as e:
            self.logger.error("appium_extended_terminal.record_video")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        return True

    @log_debug()
    def stop_video(self, **options: Any) -> Union[bytes, None]:
        """
        Останавливает запись видео. Возвращает Base64 bytes

        Возвращает:
            bool: True, если запись видео успешно остановлена, False в противном случае.
        """
        try:
            str_based64_video = self.driver.stop_recording_screen(**options)
            # Декодируем base64-кодированную строку в бинарные данные видео
            return base64.b64decode(str_based64_video)
        except NoSuchDriverException:
            self.base.reconnect()
        except InvalidSessionIdException:
            self.base.reconnect()
        except KeyError as e:
            self.logger.error("appium_extended_terminal.stop_video")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return None

    @log_debug()
    def reboot(self) -> bool:
        """
        Перезагружает устройство с помощью ADB.

        Возвращает:
            bool: True, если перезагрузка успешно запущена, False в противном случае.
        """
        try:
            self.adb_shell(command='reboot')
        except KeyError as e:
            self.logger.error("appium_extended_terminal.reboot")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)
            return False
        return True

    @log_debug()
    def get_screen_resolution(self) -> Union[Tuple[int, int], None]:
        """
        Возвращает разрешение экрана устройства с помощью ADB.

        Возвращает:
            tuple[int, int] or None: Кортеж с шириной и высотой экрана в пикселях, или None в случае ошибки.
        """
        try:
            output = self.adb_shell(command='wm', args='size')
            if "Physical size" in output:
                resolution_str = output.split(":")[1].strip()
                width, height = resolution_str.split("x")
                return int(width), int(height)
        except KeyError as e:
            self.logger.error("appium_extended_terminal.get_screen_resolution")
            self.logger.error(e)
            traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
            self.logger.error(traceback_info)

    @log_debug()
    def past_text(self, text: str, tries: int = 3) -> None:
        """
        Помещает в буфер обмена заданный текст, затем вставляет его
        """
        for _ in range(tries):
            try:
                self.driver.set_clipboard_text(text=text)
                self.input_keycode('279')
                return
            except NoSuchDriverException:
                self.base.reconnect()
            except InvalidSessionIdException:
                self.base.reconnect()

    @log_debug()
    def get_prop(self) -> dict:
        """
        Возвращает словарь атрибутов и их значений полученных командой getprop
        """
        raw_properties = self.adb_shell(command="getprop")

        # Разбиваем входную строку на строки
        lines = raw_properties.replace("\r", "").strip().split("\n")

        # Создаем пустой словарь для хранения пар ключ-значение
        result_dict = {}

        # Обрабатываем каждую строку
        for line in lines:
            try:
                # Разбиваем строку на две части, используя первое вхождение ":"
                key, value = line.strip().split(":", 1)

                # Убираем квадратные скобки вокруг ключа и значения
                key = key.strip()[1:-1]
                value = value.strip()[1:-1]
                # Добавляем пару ключ-значение в словарь
                result_dict[key] = value
            except ValueError:
                continue

        return result_dict

    def get_prop_hardware(self) -> str:
        """
        Возвращает значение атрибута 'ro.boot.hardware'
        """
        return self.get_prop()['ro.boot.hardware']

    def get_prop_model(self) -> str:
        """
        Возвращает значение атрибута 'ro.product.model'
        """
        return self.get_prop()['ro.product.model']

    def get_prop_serial(self) -> str:
        """
        Возвращает значение атрибута 'ro.serialno'
        """
        return self.get_prop()['ro.serialno']

    def get_prop_build(self) -> str:
        """
        Возвращает значение атрибута 'ro.build.description'
        """
        return self.get_prop()['ro.build.description']

    def get_prop_device(self) -> str:
        """
        Возвращает значение атрибута 'ro.product.device'
        """
        return self.get_prop()['ro.product.device']

    def get_prop_uin(self) -> str:
        """
        Вовзращает УИН устройства
        """
        return self.get_prop()['sys.atol.uin']

    def get_packages(self) -> list:
        """
        Возвращает список пакетов установленных на устройство
        """
        # Get the output from adb_shell command
        output = self.adb_shell(command='pm', args='list packages')

        # Split the output by newline to get each line separately
        lines = output.strip().split('\n')

        # Extract package names from each line and return as a list
        packages = [line.split(':')[-1].replace('\r', '') for line in lines]
        return packages

    def get_package_path(self, package: str) -> str:
        """
        Возвращает путь к пакету на устройстве.
        Args:
            package (str): Наименование пакета
        Returns:
            str: путь к пакету на устройстве
        """
        return self.adb_shell(command='pm', args=f'path {package}'). \
            replace('package:', ''). \
            replace('\r', ''). \
            replace('\n', '')

    def pull_package(self, package: str, path: str = '', filename: str = 'temp.apk'):
        """
        Скачивает с устройства указанный пакет.
        Args:
            package (str): Наименование пакета
            path (str): Путь к папке сохранения пакета на ПК
            filename (str): Имя файла сохраняемого на ПК
        """
        package_path = self.get_package_path(package=package)
        if not filename.endswith('.apk'):
            filename = f"{filename}.apk"
        self.pull(source=package_path, destination=os.path.join(path, filename))

    def get_package_manifest(self, package: str) -> dict:
        """
        Получает манифест с пакета на устройстве и возвращает его преобразованный в словарь.
        Args:
            package (str): наименование пакета
        Returns:
            dict: манифест в виде словаря
        """
        if not os.path.exists("test"):
            os.makedirs(name='test')

        self.pull_package(package=package, path="test",
                          filename="temp.apk")

        command = ["aapt", "dump", "badging", os.path.join("test", "temp.apk")]
        try:
            output: str = str(subprocess.check_output(command)).strip()
        except subprocess.CalledProcessError as error:
            # self.logger.error(f"get_package_manifest ошибка получения манифеста: {error}")
            return {}
        # Убираем лишние символы
        output = output.replace('\\r\\n', ' ').replace('b"', '').replace('"', '').replace(":'", ": '")

        # Строим список
        list_of_elements = output.split()
        result = {}
        current_key = None

        for element in list_of_elements:
            if element.endswith(':'):
                result[element] = []
                current_key = element
                continue
            result[current_key].append(element.replace("'", ""))

        os.remove(os.path.join('test', 'temp.apk'))

        return result



# coding: utf-8
"""
Этот модуль содержит полезные декораторы для работы с Appium.
"""
import base64
import sys
import io
import time
import functools
import traceback
from functools import wraps
from datetime import datetime

import allure
import numpy as np
import pytest
from PIL import Image


# TODO make unit test for module

def retry(func):
    """
    Повторяет выполнение метода если он возвращает False или None.
    3 tries hardcode in method
    """
    max_retries = 3

    @wraps(func)
    def wrapper(*args, **kwargs):
        result = None
        for _ in range(max_retries):
            result = func(*args, **kwargs)
            if result is not None and result is not False:
                return result
            time.sleep(1)
        return result

    # Возвращаем обертку функции
    return wrapper


def wait_until_window_change(poll_frequency: float = 0.1):
    """
    Декоратор, который ожидает пока окно не перестанет меняться.
    В обернутом методе должен быть аргумент
        decorator_args: dict
            timeout_window: общее время ожидания
            window_not_changing_period: период времени в течении которого окно не должно изменятся

    Аргументы:
        poll_frequency (float): Частота опроса содержимого окна на наличие изменений в секундах.
                               По умолчанию 0.1 секунды.

    Возвращает:
        function: Декорированная функция.
    """

    def func_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            """
            Оберточная функция, которая инкапсулирует декорированную функцию
            с логикой обнаружения изменений окна.

            Аргументы:
                self: Экземпляр класса, к которому принадлежит
                      декорированный метод.
                *args: Произвольное число аргументов, переданных в
                       декорированный метод.
                **kwargs: Произвольное число именованных аргументов, переданных в
                          декорированный метод.

            Возвращает:
                bool: True, если содержимое окна изменяется в течении
                      заданного периода времени, иначе False.
            """

            # Инициализация
            func_result = False
            decorator_args = kwargs.get('decorator_args', {})
            timeout_window = decorator_args.get('timeout_window', 30)
            window_not_changing_period = decorator_args.get('window_not_changing_period', 10)
            # Запись начального времени
            start_time = time.time()
            # Вызов декорированной функции и сохранение результата
            func_result = func(self, *args,
                               **kwargs)

            # Обнаружение изменений экрана с экспоненциальной задержкой
            poll_interval = poll_frequency
            # Продолжаем до достижения тайм-аута
            while time.time() - start_time < timeout_window:
                # Запускаем новый период, в течение которого окно не изменяется
                window_not_changing_period_start_time = time.time()
                # Флаг для отслеживания того, изменилось ли окно за период
                window_not_changed = True

                while (time.time() - window_not_changing_period_start_time
                       < window_not_changing_period):
                    # Делаем снимок экрана и сохраняем его в памяти
                    image_bytes = self.driver.get_screenshot_as_png()
                    # Преобразуем в оттенки серого
                    image = Image.open(io.BytesIO(image_bytes)).convert('L')
                    # Обрезаем снимок до определенной области (лево, верх, право, низ)
                    box = (50, 50, 400, 400)
                    image = image.crop(box)
                    # Ждем указанный интервал между опросами
                    time.sleep(poll_interval)
                    new_image_bytes = self.driver.get_screenshot_as_png()
                    # Преобразуем в оттенки серого
                    new_image = Image.open(io.BytesIO(new_image_bytes)).convert('L')
                    new_image = new_image.crop(box)

                    # Проверяем, отличается ли сумма значений пикселей на двух изображениях
                    if np.sum(image) != np.sum(new_image):
                        # Содержимое окна изменилось
                        window_not_changed = False
                        break

                if window_not_changed:
                    self.logger.debug("Содержимое окна не изменялось в течение периода")
                    return func_result
                # Удваиваем время ожидания для каждого опроса
                poll_interval *= 2

            return func_result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def wait_for_window_change(poll_frequency: float = 0.5):
    """
    Декоратор, который ожидает изменения окна после выполнения метода.
    Если окно не изменилось - выполняет еще попытку.
    В обернутом методе должен быть аргумент
        decorator_args: dict
            timeout_window: int, время ожидания на попытку
            tries: int, количество попыток выполнения метода

    Аргументы:
        poll_frequency (float): Частота проверки окна на изменения.

    Возвращает:
        Декоратор функции.
    """

    def func_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            """
            Оберточная функция, которая выполняет обнаружение изменения окна и
            выполнение декорированной функции.

            Аргументы:
                self: Экземпляр класса, содержащего декорированную функцию.
                *args: Переменное число аргументов.
                **kwargs: Произвольные именованные аргументы.

            Возвращает:
                Результат декорированной функции или False, если изменение окна не было обнаружено.
            """

            # Инициализация
            func_result = False
            decorator_args = kwargs.get('decorator_args', {})
            timeout_window = decorator_args.get('timeout_window', 10)
            tries = decorator_args.get('tries', 3)

            # Сделать снимок экрана и сохранить его в памяти в виде байтов
            image_bytes = self.driver.get_screenshot_as_png()
            # Открыть изображение из байтов и преобразовать его в оттенки серого
            image = Image.open(io.BytesIO(image_bytes)).convert(
                'L')

            # Обрезать снимок экрана до определенной области (лево, верх, право, низ)
            box = (50, 50, 400, 400)
            # Обрезать изображение на основе заданных координат области
            image = image.crop(box)

            # Попытаться обнаружить изменение экрана в течении tries попыток
            for _ in range(tries):
                # Записать текущее время начала попытки обнаружения
                start_time = time.time()
                # Выполнить декорированную функцию и сохранить результат
                func_result = func(self, *args,
                                   **kwargs)

                # Обнаружить изменение экрана с экспоненциальной задержкой
                poll_interval = poll_frequency
                # Проверить, находится ли прошедшее время в пределах заданного окна времени ожидания
                while time.time() - start_time < timeout_window:
                    # Приостановить выполнение на заданный интервал проверки
                    time.sleep(poll_interval)
                    # Сделать новый снимок экрана окна и получить данные изображения в виде байтов
                    new_image_bytes = self.driver.get_screenshot_as_png()
                    # Открыть новое изображение из байтов и преобразовать его в оттенки серого
                    new_image = Image.open(io.BytesIO(new_image_bytes)).convert(
                        'L')
                    # Обрезать новое изображение на основе заданных координат области
                    new_image = new_image.crop(
                        box)
                    # Сравнить суммы значений пикселей между исходным и новым изображениями
                    if not np.sum(image) == np.sum(
                            new_image):
                        # Записать сообщение о том, что произошло изменение экрана
                        self.logger.debug(
                            "Изменение экрана обнаружено")
                        # Вернуть True для обозначения обнаружения изменения экрана
                        return func_result
                        # Удвоить интервал проверки для следующей итерации
                        # (экспоненциальная задержка)
                    poll_interval *= 2

            return func_result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def wait_until_dom_change(poll_frequency: float = 0.1):
    """
    Декоратор, который ожидает пока DOM не перестанет меняться.
    В обернутом методе должен быть аргумент
        decorator_args: dict
            timeout_window: общее время ожидания
            window_not_changing_period: период времени в течении которого окно не должно изменятся

    Аргументы:
        poll_frequency (float): Частота опроса содержимого окна на наличие изменений в секундах.
                               По умолчанию 0.1 секунды.

    Возвращает:
        function: Декорированная функция.
    """

    def func_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            """
            Оберточная функция, которая инкапсулирует декорированную функцию
            с логикой обнаружения изменений DOM.

            Аргументы:
                self: Экземпляр класса, к которому принадлежит
                      декорированный метод.
                *args: Произвольное число аргументов, переданных в
                       декорированный метод.
                **kwargs: Произвольное число именованных аргументов, переданных в
                          декорированный метод.

            Возвращает:
                bool: True, если содержимое окна изменяется в течении
                      заданного периода времени, иначе False.
            """

            # Инициализация
            func_result = False
            decorator_args = kwargs.get('decorator_args', {})
            timeout_window = decorator_args.get('timeout_window', 30)
            window_not_changing_period = decorator_args.get('window_not_changing_period', 10)
            # Запись начального времени
            start_time = time.time()
            # Вызов декорированной функции и сохранение результата
            func_result = func(self, *args,
                               **kwargs)

            # Обнаружение изменений экрана с экспоненциальной задержкой
            poll_interval = poll_frequency
            # Продолжаем до достижения тайм-аута
            while time.time() - start_time < timeout_window:
                # Запускаем новый период, в течение которого окно не изменяется
                window_not_changing_period_start_time = time.time()
                # Флаг для отслеживания того, изменилось ли окно за период
                window_not_changed = True

                while (time.time() - window_not_changing_period_start_time
                       < window_not_changing_period):
                    # Делаем снимок DOM
                    page_source = self.driver.page_source
                    # Ждем указанный интервал между опросами
                    time.sleep(poll_interval)
                    new_page_source = self.driver.page_source

                    # Проверяем, отличается ли сумма значений пикселей на двух изображениях
                    if page_source != new_page_source:
                        # Содержимое окна изменилось
                        window_not_changed = False
                        break
                if window_not_changed:
                    self.logger.debug("Содержимое DOM не изменялось в течение периода")
                    return func_result
                # Удваиваем время ожидания для каждого опроса
                poll_interval *= 2

            return func_result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def wait_for_dom_change(poll_frequency: float = 0.5):
    """
    Декоратор, который ожидает изменения DOM после выполнения метода.
    Если DOM не изменилось - выполняет еще попытку.
    В обернутом методе должен быть аргумент
        decorator_args: dict
            timeout_window: int, время ожидания на попытку
            tries: int, количество попыток выполнения метода

    Аргументы:
        poll_frequency (float): Частота проверки окна на изменения.

    Возвращает:
        Декоратор функции.
    """

    def func_decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            """
            Оберточная функция, которая выполняет обнаружение изменения окна и
            выполнение декорированной функции.

            Аргументы:
                self: Экземпляр класса, содержащего декорированную функцию.
                *args: Переменное число аргументов.
                **kwargs: Произвольные именованные аргументы.

            Возвращает:
                Результат декорированной функции или False, если изменение окна не было обнаружено.
            """

            # Инициализация
            func_result = False
            decorator_args = kwargs.get('decorator_args', {})
            timeout_window = decorator_args.get('timeout_window', 10)
            tries = decorator_args.get('tries', 3)

            # Сделать снимок DOM и сохранить его в памяти
            page_source = self.driver.page_source

            # Записать текущее время начала попытки обнаружения
            start_time = time.time()
            # Выполнить декорированную функцию и сохранить результат
            func_result = func(self, *args,
                               **kwargs)

            # Обнаружить изменение экрана с экспоненциальной задержкой
            poll_interval = poll_frequency
            # Проверить, находится ли прошедшее время в пределах заданного окна времени ожидания
            while time.time() - start_time < timeout_window:
                # Приостановить выполнение на заданный интервал проверки
                time.sleep(poll_interval)
                # Сделать новый снимок DOM
                new_page_source = self.driver.page_source
                # Сравнить
                if not page_source == new_page_source:
                    # Записать сообщение о том, что произошло изменение DOM
                    self.logger.debug(
                        "Изменение DOM обнаружено")
                    # Вернуть True для обозначения обнаружения изменения DOM
                    return func_result
                    # Удвоить интервал проверки для следующей итерации
                    # (экспоненциальная задержка)
                poll_interval *= 2

            return func_result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def time_it(func):
    """
    Замеряет время выполнения метода.
    Печатает результат замера.
    """

    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time
        print(f"Execution time of {func.__name__}: {execution_time:.2f} seconds")
        return result

    # Возвращаем обертку функции
    return wrapper


def step_info(my_str):
    """
    Декоратор, который перед вызовом метода вызывает logger.info и @allure.step,
    передавая в них строковую переменную, принятую в параметрах.

    Аргументы:
        my_str (str): Строковая переменная для использования в logger.info и @allure.step.

    Возвращает:
        function: Декоратор функций.

    Пример использования:
        @my_step_info("Мой шаг")
        def my_function():
            ...
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        @allure.step(my_str)
        def wrapper(self, *args, **kwargs):
            result = None
            # Логируем информацию перед вызовом метода
            self.logger.info(my_str)
            # Получаем скриншот до вызова метода
            screenshot = self.driver.get_screenshot_as_png()
            # Генерируем временную метку для имени скриншота
            timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            # Устанавливаем имя скриншота до вызова метода
            screenshot_name_begin = f"screenshot_begin_{timestamp}.png"
            # Имя файла видеозаписи с временной меткой
            video_filename = f'screenrecord_{timestamp}.mp4'
            self.driver.start_recording_screen()
            try:
                # Выполняем исходную функцию
                result = func(self, *args, **kwargs)
            except Exception as error:
                # Если произошло исключение, прикрепляем скриншот до вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_begin,
                              attachment_type=allure.attachment_type.PNG)
                # Получаем скриншот после вызова метода
                screenshot = self.driver.get_screenshot_as_png()
                # Генерируем временную метку для имени скриншота
                timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
                # Устанавливаем имя скриншота до вызова метода
                screenshot_name_end = f"screenshot_end_{timestamp}.png"
                # Если произошло исключение, прикрепляем скриншот после вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_end,
                              attachment_type=allure.attachment_type.PNG)

                # Если произошло исключение, прикрепляем видеозапись выполнения метода к отчету
                allure.attach(base64.b64decode(self.driver.stop_recording_screen()),
                              name=video_filename,
                              attachment_type=allure.attachment_type.MP4)

                # Прикрепляем информацию об ошибке AssertionError к отчету
                allure.attach(str(error),
                              name=str(error),
                              attachment_type=allure.attachment_type.TEXT)

                # Выводим информацию в лог
                self.logger.error(f"{my_str} [не выполнено]")
                traceback_info = "".join(traceback.format_tb(sys.exc_info()[2]))
                error_msg = f"""Ошибка: {error},
                                        {args=},
                                        {kwargs=},
                            Traceback:
                            {traceback_info=}
                                """
                self.logger.error(error_msg)

                # В случае исключения помечаем тест провалившимся
                try:
                    pytest.fail(f"{func.__name__}({args}, {kwargs}), {error}")
                except Exception as e:
                    self.logger.error("Pytest не обнаружен")
                    raise

            # Логируем информацию после успешного выполнения метода
            self.logger.info(f"{my_str} [выполнено успешно]")
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def screenshots():
    """
    В случае возникновения AssertionError в обернутом методе -
    прикрепляет к allure report скриншот до выполнения
    метода и после возникновения исключения, а также информацию об ошибке.
    В ином случае скриншот не прикрепляется.
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            # Получаем скриншот до вызова метода
            screenshot = self.driver.get_screenshot_as_png()
            # Генерируем временную метку для имени скриншота
            timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            # Устанавливаем имя скриншота до вызова метода
            screenshot_name_begin = f"screenshot_begin_{timestamp}.png"
            try:
                # Выполняем исходную функцию
                result = func(self, *args, **kwargs)
            except AssertionError as error:
                # Если произошло исключение, прикрепляем скриншот до вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_begin,
                              attachment_type=allure.attachment_type.PNG)
                # Прикрепляем информацию об ошибке AssertionError к отчету
                allure.attach(str(error),
                              name="AssertionError",
                              attachment_type=allure.attachment_type.TEXT)
                # Рейзим исключение AssertionError с сохраненным traceback
                raise AssertionError(str(error)).with_traceback(sys.exc_info()[2])
            finally:
                # Получаем скриншот после вызова метода
                screenshot = self.driver.get_screenshot_as_png()
                # Обновляем временную метку для имени скриншота
                timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
                # Устанавливаем имя скриншота после вызова метода
                screenshot_name_end = f"screenshot_end_{timestamp}.png"
                # Прикрепляем скриншот после вызова метода к отчету
                allure.attach(screenshot,
                              name=screenshot_name_end,
                              attachment_type=allure.attachment_type.PNG)
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def log_debug():
    """
    Логирует начало и завершение обернутого метода
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            # Получаем имя метода
            method_name = func.__name__
            # Логируем начало выполнения метода и переданные аргументы
            self.logger.debug(f"{method_name}() < {', '.join(map(str, args))}, "
                              f"{', '.join(f'{k}={v}' for k, v in kwargs.items())}")
            # Выполняем исходную функцию
            result = func(self, *args, **kwargs)
            # Если результат существует, логируем его
            if result:
                self.logger.debug(f"{method_name}() > {str(result)}")
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator


def print_me():
    """
    Печатает начало и завершение обернутого метода
    """

    # Определяем декоратор функций
    def func_decorator(func):
        # Создаем обертку функции, сохраняющую метаданные исходной функции
        def wrapper(*args, **kwargs):
            # Получаем имя метода
            method_name = func.__name__
            # Печатаем начало выполнения метода и переданные аргументы
            print(f"{method_name}() < {', '.join(map(str, args))}, "
                  f"{', '.join(f'{k}={v}' for k, v in kwargs.items())}")
            # Выполняем исходную функцию
            result = func(*args, **kwargs)
            # Если результат существует, логируем его
            if result:
                print(f"{method_name}() > {str(result)}")
            # Возвращаем результат выполнения исходной функции
            return result

        # Возвращаем обертку функции
        return wrapper

    # Возвращаем декоратор функций
    return func_decorator
